package animation;

import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreePath;

/**
 * represents a line of pseudocode
 * @author kolja
 *
 */
public class PseudoCodeNode extends DefaultMutableTreeNode {

    public static enum CodeAction
    {
        SKIP,
        STOP,
        CONTINUE
    }

    public static enum CodeStatus
    {
        UNFINISHED,
        BREAKPOINT,
        FINISHED
    }
    
    private static final long serialVersionUID = 1L;
    
    private boolean selected;
    private JTree tree;
    private boolean breakPoint;
    private AnimationController controller;
    private int currentCodeLine; // next forward code line
    private int beforeCodeLine; // next backwards code line
    private CodeLine code;
    private boolean loop;
    private boolean stackFrameAdded;
    
    public PseudoCodeNode( String description, JTree tree, CodeLine line )
    {
        super( description );
        selected = false;
        this.tree = tree;
        breakPoint = false;
        currentCodeLine = -1;
        beforeCodeLine = -2;
        code = line;
        loop = false;
        stackFrameAdded = false;
    }
    
    public void setController( AnimationController c )
    {
        if( children != null )
        {
            for( Object ch : children )
            {
                ((PseudoCodeNode)ch).setController( c );
            }
        }
        controller = c;
    }
    
    @Override
    public void add( MutableTreeNode node )
    {
        ((PseudoCodeNode)node).setController( controller );
        super.add( node );
    }
    
    /**
     * 
     * @return the tree that this node belongs to
     */
    public JTree getTree()
    {
        return tree;
    }
    
    /**
     * checks if this node should be highlighted
     * @return true if it should, false otherwise
     */
    public boolean isSelected()
    {
        return selected;
    }
    
    /**
     * checks if one of the subnodes of this node is selected.
     * @return true if one is, false otherwise
     */
    public boolean hasSelectedSubnode()
    {
    	if( children != null )
    	{
	    	for( Object ch : children )
	        {
	            if( ((PseudoCodeNode)ch).isSelected() || ((PseudoCodeNode)ch).hasSelectedSubnode() )
	            	return true;
	        }
    	}
    	return false;
    }
    
    /**
     * highlight this line of pseudocode.
     * should be called when the line is entered, as it triggers breakpoints
     * @param selected whether to select or deselect this line
     * @return false iff a breakpoint was reached and the node was set to be selected.
     */
    public CodeAction setSelected( boolean selected )
    {
        if( selected && breakPoint )
            controller.setContinuous( false );
        this.selected = selected;
        if( selected )
        {
            if( controller == null || controller.getStepOption() != 1 || breakPoint )
                tree.expandPath( new TreePath( this.getPath() ) );
        }
        else
        {
            if( controller == null || controller.getStepOption() != 1 )
                tree.collapsePath( new TreePath( this.getPath() ) );
        }
        if( breakPoint && selected )
            return CodeAction.STOP; // Breakpoint
        if( controller != null && controller.getStepOption() == 1 && !tree.isVisible( new TreePath( this.getPath() ) ) )
            return CodeAction.SKIP; // Step would be to detailed
        return CodeAction.CONTINUE; // Normal
    }
    
    /**
     * set a breakpoint at this line of code
     * @param breakPoint whether there should be a breakpoint or node
     */
    public void setBreakPoint( boolean breakPoint )
    {
        this.breakPoint = breakPoint;
    }
    
    /**
     * check if there is a breakpoint set at this line of code
     * @return true, iff there is a breakpoint
     */
    public boolean hasBreakPoint()
    {
        return breakPoint;
    }
    
    private PseudoCodeNode getForwardNode()
    {
        if( currentCodeLine == -1 )
            return this;
        if( children != null && children.size() > currentCodeLine )
            return ((PseudoCodeNode)children.get( currentCodeLine )).getForwardNode();
        return this;
    }
    
    private PseudoCodeNode getBackwardNode()
    {
        if( beforeCodeLine == -1 )
            return this;
        if( children != null && children.size() > beforeCodeLine )
            return ((PseudoCodeNode)children.get( beforeCodeLine )).getBackwardNode();
        return this;
    }
    
    private CodeStatus stepInto( Memory m )
    {
        currentCodeLine = 1;
        beforeCodeLine = 0;
        stackFrameAdded = true;
        if( children == null || children.size() == 0 )
        {
            setSelected( false );
            if( stackFrameAdded )
                m.removeFrame();
            stackFrameAdded = false;
            return CodeStatus.FINISHED;
        }
        else
        {
            setSelected( false );
            CodeStatus status = selectChild( 0, m ); // TODO if loop do something when finished
            if( status == CodeStatus.FINISHED )
            {
                if( loop )
                {
                    beforeCodeLine = currentCodeLine;
                    currentCodeLine = -1; // TODO select this node
                    status = CodeStatus.UNFINISHED;
                }
                else
                {
                    if( stackFrameAdded )
                        m.removeFrame();
                    stackFrameAdded = false;
                }
            }
            return status;
        }
    }
    
    /**
     * Perform one atomic step of the algorithm. Stops at the end of the program.
     * @return whether the whole stage is finished (afterwards).
     * For example if all steps are finished, then {@code FINISHED} is returned.
     */
    public CodeStatus forwardStep( Memory m )
    {
        if( currentCodeLine == -1 )
        {
            ControlFlow cf = code.runForward( m );
            switch( cf.getStatus() )
            {
            case ControlFlow.STEP_INTO_LOOP:
                loop = true;
                return stepInto( m );
            case ControlFlow.STEP_INTO_STACK:
                stackFrameAdded = true;
                return stepInto( m );
            case ControlFlow.STEP_INTO:
                return stepInto( m );
            case ControlFlow.STEP_OVER:
                if( stackFrameAdded )
                    m.removeFrame();
                stackFrameAdded = false;
                return CodeStatus.FINISHED;
            }
        }
        else
        {
            if( children == null || children.size() <= currentCodeLine )
            {
                if( stackFrameAdded )
                    m.removeFrame();
                return CodeStatus.FINISHED;
            }
            switch( ( (PseudoCodeNode)children.get( currentCodeLine ) ).forwardStep( m ) )
            {
            case BREAKPOINT:
                return CodeStatus.BREAKPOINT;
            case FINISHED:
                ( (PseudoCodeNode)children.get( currentCodeLine ) ).setSelected( false );
                beforeCodeLine = currentCodeLine;
                currentCodeLine++;
                if( children.size() <= currentCodeLine )
                {
                    if( loop )
                    {
                        currentCodeLine = -1; // TODO select this node
                    }
                    else
                    {
                        if( stackFrameAdded )
                            m.removeFrame();
                        return CodeStatus.FINISHED;
                    }
                }
                else
                {
                    CodeStatus status = selectChild( currentCodeLine, m );
                    if( status == CodeStatus.FINISHED )
                    {
                        if( loop )
                        {
                            beforeCodeLine = currentCodeLine;
                            currentCodeLine = -1; // TODO select this node
                            status = CodeStatus.UNFINISHED;
                        }
                        else
                        {
                            if( stackFrameAdded )
                                m.removeFrame();
                        }
                    }
                    return status;
                }
            case UNFINISHED:
                return CodeStatus.UNFINISHED;
            }
        }
        return CodeStatus.UNFINISHED;
    }
    
    private CodeStatus selectChild( int index, Memory m  )
    {
        switch( ( (PseudoCodeNode)children.get( index ) ).setSelected( true ) )
        {
        case CONTINUE:
            return CodeStatus.UNFINISHED;
        case SKIP:
            switch( ( (PseudoCodeNode)children.get( index ) ).forwardStepOverIntern( m ) )
            {
            case BREAKPOINT:
                return CodeStatus.BREAKPOINT;
            case FINISHED:
                ( (PseudoCodeNode)children.get( index ) ).setSelected( false );
                beforeCodeLine = currentCodeLine;
                currentCodeLine++;
                if( children == null || currentCodeLine >= children.size() )
                {
                    return CodeStatus.FINISHED;
                }
                else
                {
                    return selectChild( currentCodeLine, m );
                }
            case UNFINISHED:
                break;
            }
        case STOP:
            return CodeStatus.BREAKPOINT;
        }
        return CodeStatus.UNFINISHED;
    }

    /**
     * Perform steps until the next line of code on the same level of indentation as this line
     * is reached. Stops at the end of the program.
     * @return whether the whole stage is finished (afterwards).
     * For example if all steps are finished, then {@code FINISHED} is returned.
     */
    public CodeStatus forwardStepOver( Memory m )
    {
        return getForwardNode().forwardStepOverIntern( m );
    }
    
    private CodeStatus forwardStepOverIntern( Memory m )
    {
        CodeStatus status = null;
        do {
            status = forwardStep( m );
        } while( status == CodeStatus.UNFINISHED );
        return status;
    }

    /**
     * Perform steps until the next line of code on the level of indentation above this lines 
     * level is reached. Stops at the end of the program.
     * @return whether the whole stage is finished (afterwards).
     * For example if all steps are finished, then {@code FINISHED} is returned.
     */
    public CodeStatus forwardStepOut( Memory m )
    {
        return getForwardNode().forwardStepOutIntern( m );
    }
    
    private CodeStatus forwardStepOutIntern( Memory m )
    {
        if( parent != null )
            return ((PseudoCodeNode)parent).forwardStepOverIntern( m );
        return forwardStepOverIntern( m );
    }
    
    /**
     * Undo one atomic step of the algorithm. Stops at the beginning of the program.
     * @return whether the whole stage is finished in backwards direction (afterwards).
     * For example if all steps have been reverted, then {@code FINISHED} is returned.
     */
    public CodeStatus backwardStep( Memory m )
    {
        // TODO
        return null;
    }

    /**
     * Perform backward steps until the previous line of code on the same level of indentation 
     * as this line is reached. Stops at the end of the program.
     * @return whether the whole stage is finished in backwards direction (afterwards).
     * For example if all steps have been reverted, then {@code FINISHED} is returned.
     */
    public CodeStatus backwardStepOver( Memory m )
    {
        return getBackwardNode().backwardStepOverIntern( m );
    }
    
    private CodeStatus backwardStepOverIntern( Memory m )
    {
        CodeStatus status = null;
        do {
            status = backwardStep( m );
        } while( status == CodeStatus.UNFINISHED );
        return status;
    }


    /**
     * Perform backward steps until the previous line of code on the level of indentation above 
     * this lines level is reached. Stops at the end of the program.
     * @return whether the whole stage is finished in backwards direction (afterwards).
     * For example if all steps have been reverted, then {@code FINISHED} is returned.
     */
    public CodeStatus backwardStepOut( Memory m )
    {
        return getBackwardNode().backwardStepOutIntern( m );
    }
    
    private CodeStatus backwardStepOutIntern( Memory m )
    {
        if( parent != null )
            return ((PseudoCodeNode)parent).backwardStepOver( m );
        return backwardStepOutIntern( m );
    }
}
